@@ -8,6 +8,31 @@ import networkx as nx
 from skspatial.objects import Line
 from scipy.spatial import distance
 
+def nx_to_Graph(G, Weight = False):
+    """
+    Description
+    -----------
+    Performs a conversion from the **networkx** graph format to the **fastnode2vec** one, that is necessary to work with
+    fastnode2vec objects.
+
+    Parameters
+    ----------
+    G : networkx.Graph()
+        Gives the network that will be converted.
+    Weight : bool
+        Specifies if the network is a weighted one.
+
+    Returns
+    -------
+    output : fastnode2vec.Graph
+        The output of the function is a **fastnode2vec** Graph object.
+    """
+    if Weight == False:
+        G_fn2v = Graph(G.edges(), directed = False, weighted = Weight)
+    else:
+        G_fn2v = Graph(list(G.edges.data("weight", default = 1)), directed = False, weighted = Weight)
+    return G_fn2v
+
 def generate_edgelist(df):
     """
     Description
@@ -40,12 +65,12 @@ def generate_edgelist(df):
     >>> edgelist = xn2v.generate_edgelist(df)
         [('1.0', '2.0', 3.7), ('1.0', '3.0', 0.33), ('2.0', '7.0', 12.0)]
     """
-    # forcing type values
-    df = df.astype({'node1': str, 'node2': str, 'weight': np.float64})
     # check header:
     header_names = list(df.columns.values)
     if header_names[0] != 'node1' or header_names[1] != 'node2' or header_names[2] != 'weight':
         raise TypeError('The header format is different from the required one.')
+    # forcing type values
+    df = df.astype({'node1': str, 'node2': str, 'weight': np.float64})
     return list(df.itertuples(index = False, name = None))
 
 def edgelist_from_csv(path, **kwargs):
@@ -73,7 +98,7 @@ def edgelist_from_csv(path, **kwargs):
     >>> G = nx.Graph()
     >>> G.add_weighted_edges_from(edgelist)
     - The data types of the 'node1' and 'node2' columns must be strings, otherwise they will be converted as strings.
-    
+
     Examples
     --------
     >>> edgelist = xn2v.edgelist_from_csv('somefile.csv')
@@ -113,29 +138,28 @@ def complete_edgelist(Z, info=False, **kwargs):
         >>> points = np.column_stack((x1, y1))
         >>> df = xn2v.complete_edgelist(points)
               node1 node2    weight
-            0     0     0  0.000000
-            1     0     1  1.358972
-            2     0     2  2.393888
-            3     1     0  1.358972
-            4     1     1  0.000000
-            5     1     2  1.345274
-            6     2     0  2.393888
-            7     2     1  1.345274
-            8     2     2  0.000000
+            0     0     0  1.000000
+            1     0     1  0.015445
+            2     0     2  0.018235
+            3     1     0  0.015445
+            4     1     1  1.000000
+            5     1     2  0.834821
+            6     2     0  0.018235
+            7     2     1  0.834821
+            8     2     2  1.000000
     """
     dimension = Z[0].size  # Number of coordinates per point
     NPoints = Z[:, 0].size  # Number of points
-    weights = distance.cdist(Z, Z, 'euclidean') # Distance between all points
-    df = pd.DataFrame(columns = ['node1', 'node2', 'weight'], **kwargs)
-    l = 0
-    for i in range(0, NPoints):
-        for j in range(0, NPoints):
-            df.loc[l] = [f"{i}", f"{j}", weights[i][j]]
-            l+=1
+    weights = np.exp(-distance.cdist(Z, Z, 'euclidean'))  # Distance between all points
+    weights = weights.flatten() # Weights coulumn
+    nodes_id = np.arange(NPoints).astype(str)
+    node1 = np.repeat(nodes_id,NPoints)
+    node2 = np.tile(nodes_id,NPoints)
+    df = pd.DataFrame({'node1': node1, 'node2': node2, 'weight': weights}, **kwargs)
     if info == True:
         print('\033[1m' + '--------- General Information ---------')
         print('Edge list of a fully connected network.')
-        print('The weights are calculated using the euclidean norm.\n')
+        print('The weights are calculated using minus the exponential of the euclidean norm.\n')
         print('- Space dimensionality: ', dimension)
         print('- Number of Points: ', NPoints)
         print('- Minimum weight: ', np.min(weights))
@@ -170,40 +194,40 @@ def stellar_edgelist(Z, info=False, **kwargs):
     >>> y1 = np.random.normal(9, 1, 6)
     >>> points_1 = np.column_stack((x1, y1))
     >>> df = xn2v.stellar_edgelist(points_1)
-          node1 node2     weight
-        0     0     1  12.571278
-        1     0     2  11.765633
-        2     0     3   9.735974
-        3     0     4  12.181443
-        4     0     5  11.027584
-        5     0     6  12.755861
-
+          node1      node2     weight
+        0     origin     0  12.571278
+        1     origin     1  11.765633
+        2     origin     2   9.735974
+        3     origin     3  12.181443
+        4     origin     4  11.027584
+        5     origin     5  12.755861
     >>> x2 = np.random.normal(107, 2, 3)
     >>> y2 = np.random.normal(101, 1, 3)
     >>> points_2 = np.column_stack((x2, y2))
     >>> tot = np.concatenate((points_1,points_2),axis=0)
     >>> df = xn2v.stellar_edgelist(tot)
-          node1 node2     weight
-        0     0     1  12.571278
-        1     0     2  11.765633
-        2     0     3   9.735974
-        3     0     4  12.181443
-        4     0     5  11.027584
-        5     0     6  12.755861
-        6     0     7  146.229997
-        7     0     8  146.952899
-        8     0     9  146.595700
+          node1      node2     weight
+        0     origin     0  12.571278
+        1     origin     1  11.765633
+        2     origin     2   9.735974
+        3     origin     3  12.181443
+        4     origin     4  11.027584
+        5     origin     5  12.755861
+        6     origin     6  146.229997
+        7     origin     7  146.952899
+        8     origin     8  146.595700
     """
-    dimension = Z[0].size # Number of coordinates per point
-    NPoints = Z[:,0].size # Number of points
-    weights = np.linalg.norm(Z, axis = 1)
-    df = pd.DataFrame(columns = ['node1', 'node2', 'weight'], **kwargs)
-    for i in range(0,NPoints):
-        df.loc[i] = ['0',f"{i+1}",weights[i]]
+    dimension = Z[0].size  # Number of coordinates per point
+    NPoints = Z[:, 0].size  # Number of points
+    dimension = Z[0].size  # Number of coordinates per point
+    NPoints = Z[:, 0].size  # Number of points
+    weights = np.exp(-np.linalg.norm(Z, axis = 1))
+    node2 = np.arange(NPoints).astype(str)
+    df = pd.DataFrame({'node1': 'origin', 'node2': node2, 'weight': weights}, **kwargs)
     if info == True:
-        print('\033[1m'+'--------- General Information ---------')
+        print('\033[1m' + '--------- General Information ---------')
         print('Edge list of a stellar network.')
-        print('The weights are calculated using the euclidean norm.\n')
+        print('The weights are calculated using minus the exponential of the euclidean norm.\n')
         print('- Space dimensionality: ', dimension)
         print('- Number of Points: ', NPoints)
         print('- Minimum weight: ', np.min(weights))
@@ -212,6 +236,41 @@ def stellar_edgelist(Z, info=False, **kwargs):
         print('- Weight Variance: ', np.var(weights))
     return df
 
+def low_limit_network(G,delta=1.):
+    """
+    Description
+    -----------
+    This function performs a **network transformation**. It sets the link weights of the network to 0 if their initial
+    value was below a given threshold. The threshold is chosen to be a constant times the average links weight.
+    
+    Parameters
+    ----------
+    G : networkx.Graph()
+        Gives the network that will be modified.
+    delta :  float, optional
+        Set the multiplying constant of the average link weight that will define the weight threshold.
+
+    Returns
+    -------
+    output : networkx.Graph()
+        Returns the networkx graph() resulting after the transformation.
+
+    Note
+    ----
+    - The number of nodes and edges of the original network won't change. Only specific weight values will be set to 0.
+    
+    Examples
+    --------
+    >>> G = xn2v.low_limit_network(G,1.9)
+    """
+    link_weights = nx.get_edge_attributes(G, 'weight')
+    weights = np.array(list(link_weights.values())).astype(float)
+    average_weight = np.mean(weights)
+    for u, v, d in G.edges(data = True):
+        if d['weight'] < delta * average_weight:
+            d['weight'] = 0.
+    return G
+
 def best_line_projection(Z):
     """
     Description
@@ -223,12 +282,12 @@ def best_line_projection(Z):
     Z : numpy ndarray
         Numpy array containing as columns the i-th coordinate of the k-th point. The rows are the points, the columns
         are the coordinates.
-
+        
     Returns
     -------
     output : numpy ndarray
         The output of the function is a numpy ndarray containing the transformed points of the dataset.
-
+        
     Examples
     --------
     >>> x1 = np.random.normal(7, 1, 6)
@@ -243,11 +302,145 @@ def best_line_projection(Z):
     NPoints = Z[:, 0].size
     dimension = Z[0].size
     projections = []
-    for i in range(0,NPoints):
+    for i in range(0, NPoints):
         projections.extend(np.array(a.project_point(Z[i])))
     projections = np.reshape(projections, (NPoints, dimension))
     return projections
 
+def cluster_generation(result, cluster_rigidity = 0.7):
+    """
+        Description
+        -----------
+        This function takes the nodes that have a similarity higher than the one set by *cluster_rigidity*.
+
+        Parameters
+        ----------
+        result : list
+            This parameter must be a list of two elements: the first is the nodes labels vector, the other is their
+            similarities vector.
+        cluster_rigidity : float, optional
+            Sets the similarity threshold of the nodes. It should be a number between 0 and 1. The default value is
+            '0.7'.
+
+        Returns
+        -------
+        output : numpy ndarray
+            The output is a numpy array containing the nodes that satisfy the condition required.
+
+        Examples
+        --------
+        >>> nodes = np.arange(0,5)
+            [0 1 2 3 4]
+        >>> similarities = [0.5,0.9,0.91,0.87,0.67]
+        >>> xn2v.cluster_generation([nodes,similarities], cluster_rigidity = 0.75)
+            ['1' '2' '3']
+    """
+    cluster = np.array(result[0],dtype = str)
+    positions = np.where(np.array(result[1]) >= cluster_rigidity)
+    return(cluster[positions])
+
+def clusters_detection(G, cluster_rigidity=0.7, spacing=5, dim_fraction=0.8, **kwargs):
+    """
+        Description
+        -----------
+        This function detects the **clusters** that compose a generic dataset. The dataset must be given as a
+        **networkx** graph, using the proper data transformation. The clustering procedure uses Node2Vec algorithm to
+        find the most similar nodes in the network.
+
+        Parameters
+        ----------
+        G : networkx.Graph()
+            Gives the network that will be analyzed.
+        cluster_rigidity : float, optional
+            Sets the similarity threshold of the nodes. It should be a number between 0 and 1. The default value is
+            '0.7'.
+        spacing : int, optional
+            Sets the increment of the position when picking the nodes for the analysis. The nodes are picked every
+            *spacing* value between the initial node label and the last one. The default value is '5'.
+        dim_fraction : float, optional
+            Sets the minumum threshold for choosing when to expand the previously generated cluster. This parameter
+            appears when considering the current most similar nodes picked by the Node2Vec algorithm; in particular
+            it sets the minumum number of nodes that are already present in the generated clusters family before adding
+            the remaining ones to the previous cluster.
+
+        Note
+        ----
+        - The function returns only the nodes labels from a specific network. In order to go back to the initial points
+          it's necessary to use the function **xn2v.recover_points(dataset,Graph,Clusters[i])**. Up to now, you can
+          insert only a one-dimensional vector inside **xn2v.recover_points()**, meaning that if you have N clusters
+          inside the list obtained by **xn2v.clusters_detection()** you'll have to call **xn2v.recover_points()**
+          N times.
+
+        Returns
+        -------
+        output : list
+            The output is list of numpy arrays containing nodes labels from the given network *G*. Each array represents
+            a specific cluster.
+
+        Examples
+        --------
+        >>> df = xn2v.edgelist_from_csv('somefile.csv')
+        >>> df = xn2v.generate_edgelist(df)
+        >>> xn2v.clusters_detection(G, cluster_rigidity = 0.75, spacing = 5, dim_fraction = 0.8, picked=100,
+        >>>                         dim=100,context=5,Weight=True, walk_length=20)
+            --------- Clusters Information ---------
+            - Number of Clusters:  3
+            - Total nodes:  61
+            - Clustered nodes:  52
+            - Number of unlabeled nodes:  9
+            [array(['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19',
+            '2', '3', '4', '5', '6', '7', '9'], dtype='<U2'),
+            array(['21', '23', '24', '26', '27', '28', '29', '30', '31', '32', '33',
+            '34', '35', '36', '37', '38', '39'], dtype='<U2'),
+            array(['41', '42', '44', '45', '47', '48', '49', '50', '51', '52', '53',
+            '54', '55', '56', '57', '58', '59'], dtype='<U2')]
+    """
+    clusters = []
+    nodes_clustered = []
+    index = 0
+    for node_id in list(G.nodes)[::spacing]:
+        nodes, similarities = similar_nodes(G, node_id, **kwargs)
+        current_cluster = cluster_generation([nodes, similarities], cluster_rigidity) # TESTING THIS OBJECT.
+        dimension = np.size(current_cluster)
+        different_nodes_counters = [] # This is used to find the cluster to expand and to decide whether to create a new cluster.
+        if len(clusters) != 0:
+            for previous_cluster in clusters:
+                # True = already in, False = new node. This is inverted, so True = new node.
+                current_positions = np.invert(np.isin(current_cluster, previous_cluster))
+                different_nodes_counters.append(current_positions.sum())
+                current_cluster = current_cluster[~np.in1d(current_cluster,previous_cluster)] # current_cluster filtering.
+        if dimension == np.size(current_cluster):
+            # Creating new cluster if the dimension of cluster remain the same. This means that the nodes in common are none.
+            if dimension == 0: raise Exception("Error: The dimension of the cluster is 0. Try to reduce cluster_rigidity.")
+            print("- Creating new cluster")
+            clusters.append(current_cluster)
+            index += 1
+        elif dimension - np.size(current_cluster) < int(dimension * dim_fraction):
+            # - If all nodes are different, dimension-np.size(current_cluster)==0 => Create new cluster.
+            # - If all nodes are the same, dimension-np.size(current_cluster)==dimension => Expand with nothing.
+            # - If some nodes are different, dimension-np.size(current_cluster)>0:
+            #       if dimension-np.size(current_cluster)<threshold => There are enough SAME nodes to expand the clusters.
+            ind = np.where(different_nodes_counters == np.min(different_nodes_counters))
+            if ind[0].size != 1: # Don't expand if there are ambiguities on where put the nodes.
+                pass
+            else:
+                # We extend the cluster that had the highest number of nodes in common. The current_cluster array is
+                # already filtered with the different nodes in the whole clusters list.
+                print(f"- Expand cluster number: {ind[0][0]+1}")
+                clusters[ind[0][0]] = np.append(clusters[ind[0][0]], current_cluster)
+        else:
+            pass
+    tot_nodes = [val for sublist in clusters for val in sublist]  # Flatten list of lists
+    unlabeled = list(set(list(G.nodes)) - set(tot_nodes))  # Get remaining nodes
+    print('\033[1m' + '--------- Clusters Information ---------')
+    print('- Number of Clusters: ', index)
+    print('- Total nodes: ', len(tot_nodes) + len(unlabeled))
+    print('- Clustered nodes: ', len(tot_nodes))
+    print('- Number of unlabeled nodes: ', len(unlabeled))
+    for i in range(0,len(clusters)):
+        print(f"- Nodes in cluster {i+1}:", clusters[i].size)
+    return clusters, unlabeled
+
 def recover_points(Z, G, nodes):
     """
         Description
@@ -296,13 +489,15 @@ def recover_points(Z, G, nodes):
          [16.30640697 12.15702448]
          [18.73718742 13.99351914]
          [18.7817838   7.92318885]
-         [16.15456589 10.72636297]]       
+         [16.15456589 10.72636297]]
         """
     # check dimensionality
-    if Z[:,0].size != np.array(G.nodes).size:
-        print('dataset: ',Z[:,0].size)
-        print('network: ', np.array(G.nodes).size)
-        raise Exception('Error: the dataset dimension is different from the one of the network.')
+    if np.array(G.nodes)[0] == 'origin':
+        if Z[:,0].size+1 != np.array(G.nodes).size:
+            raise Exception(f"Error: the dataset dimension dim={Z[:, 0].size} is different from the one expected for the network dim={np.array(G.nodes).size}.")
+        else: Z = np.insert(Z, 0, np.zeros(Z[0].size), axis=0) # Adding origin
+    elif Z[:,0].size != np.array(G.nodes).size:
+            raise Exception(f"Error: the dataset dimension dim={Z[:, 0].size} is different from the one of the network dim={np.array(G.nodes).size}.")
     # force string type
     nodes = [str(s) for s in nodes]
     picked_nodes = []
@@ -313,7 +508,6 @@ def recover_points(Z, G, nodes):
         pos += 1
     return np.array(picked_nodes)
 
-
 def similar_nodes(G, node=1, picked=10, train_time = 30, Weight=False, save_model = False, 
                   model_name = 'model.wordvectors' , **kwargs):
     """
@@ -368,6 +562,7 @@ def similar_nodes(G, node=1, picked=10, train_time = 30, Weight=False, save_mode
         relatable -- belonging to the same topic. If the value is small, 2-15, then we will likely have interchangeable
         words, while if it is large, >15, we will have relatable words.
         The default value is '10'
+
     Returns
     -------
     output : ndarray, ndarray
@@ -385,16 +580,13 @@ def similar_nodes(G, node=1, picked=10, train_time = 30, Weight=False, save_mode
     Examples
     --------
     >>> G = nx.generators.balanced_tree(r=3, h=4)
-    >>> nodes, similarity = similar_nodes(G, dim=128, walk_length=30, context=10, 
+    >>> nodes, similarity = xn2v.similar_nodes(G, dim=128, walk_length=30, context=10, 
     >>>                                   p=0.1, q=0.9, workers=4)
         nodes: [0 4 5 6 45 40 14 43 13 64]
         similarity: [0.81231129 0.81083304 0.760795 0.7228986 0.66750246 0.64997339 
                      0.64365959 0.64236712 0.63170493 0.63144475]
     """
-    if Weight == False:
-        G_fn2v = Graph(G.edges(), directed = False, weighted = Weight)
-    else:
-        G_fn2v = Graph(list(G.edges.data("weight", default = 1)), directed = False, weighted = Weight)
+    G_fn2v = nx_to_Graph(G, Weight)
     n2v = Node2Vec(G_fn2v, **kwargs)
     n2v.train(epochs=train_time)
     if save_model == True:
@@ -406,7 +598,7 @@ def similar_nodes(G, node=1, picked=10, train_time = 30, Weight=False, save_mode
     similarity = np.array(similarity)
     return nodes_id, similarity
     
-def Load(file):
+def load_model(file):
     """
     Parameters
     ----------
@@ -432,41 +624,41 @@ def Load(file):
     model = Word2Vec.load(file)
     return model
 
-def Draw(G, nodes_result, title = 'Community Network', **kwargs):
+def draw_community(G, nodes_result, title='Community Network', **kwargs):
     """
-        Description
-        -----------
-        Draws a networkx plot highlighting some specific nodes in that network. The last node is higlighted in red, the
-        remaining nodes in "nodes_result" are in blue, while the rest of the network is green.
+    Description
+    -----------
+    Draws a networkx plot highlighting some specific nodes in that network. The last node is higlighted in red, the
+    remaining nodes in "nodes_result" are in blue, while the rest of the network is green.
 
-        Parameters
-        ----------
-        G : networkx.Graph object
-            Sets the network that will be drawn.
-        nodes_result : ndarray
-            Gives the nodes that will be highlighted in the network. The last element will be red, the others blue.
-        title : string, optional
-            Sets the title of the plot.
+    Parameters
+    ----------
+    G : networkx.Graph object
+        Sets the network that will be drawn.
+    nodes_result : ndarray
+        Gives the nodes that will be highlighted in the network. The last element will be red, the others blue.
+    title : string, optional
+        Sets the title of the plot.
 
-        Notes
-        -----
-        - This function returns a networkx draw plot, which is good only for networks with few nodes (~40). For larger
-          networks I suggest to use other visualization methods, like Gephi.
+    Notes
+    -----
+    - This function returns a networkx draw plot, which is good only for networks with few nodes (~40). For larger
+      networks I suggest to use other visualization methods, like Gephi.
 
-        Examples
-        --------
-        >>> G = nx.generators.balanced_tree(r=3, h=4)
-        >>> nodes, similarity = similar_nodes(G, dim=128, walk_length=30, context=100, 
-        >>>                                   p=0.1, q=0.9, workers=4)
-        >>> red_node = 2
-        >>> nodes = np.append(nodes, red_node)
-        >>> Draw(G, nodes)
+    Examples
+    --------
+    >>> G = nx.generators.balanced_tree(r=3, h=4)
+    >>> nodes, similarity = xn2v.similar_nodes(G, dim=128, walk_length=30, context=100, 
+    >>>                                   p=0.1, q=0.9, workers=4)
+    >>> red_node = 2
+    >>> nodes = np.append(nodes, red_node)
+    >>> xn2v.draw_community(G, nodes)
     """
     color_map = []
     for node in G:
-        if node == int(nodes_result[-1]):
+        if str(node) == str(nodes_result[-1]):
             color_map.append('red')
-        elif node in nodes_result:
+        elif str(node) in nodes_result.astype(str):
             color_map.append('blue')
         else:
             color_map.append('green')
